 /*
  * Unitex
  *
  * Copyright (C) 2001-2006 Universit de Marne-la-Valle <unitex@univ-mlv.fr>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  * 
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
  *
  */

#include "Keyboard.h"
#include "Error.h"
#include "CharNames.h"

/*
 This library deals with non-special keys of keyboards.
 NOTE: we do not take into account dead keys, i.e. keys that
 must be combined with other keys to produce a symbol.
 Ex: the dead circumflex key must be combined with a letter to 
     produce a letter, whereas the normal circumflex key prints
     the '^' char.

 As an example, here is the fr-latin1 keyboard:

      Keyboard name=fr-latin1

+------------------------------------------------------------------------+
     1   2   3   4   5   6   7   8   9   0      +   <---  
    &    ~ " # ' { ( [ - |    _ \  ^  @ ) ] = }       
+------------------------------------------------------------------------
 |<-   A   Z   E   R   T   Y   U   I   O   P             
 ->|   a   z   e  r   t   y   u   i   o   p       $  <-+ 
+--------------------------------------------------------------------+   
 Caps    Q   S   D   F   G   H   J   K   L   M   %        
 Lock    q   s   d   f   g   h   j   k   l   m      *     
+------------------------------------------------------------------------
       >   W   X   C   V   B   N   ?   .   /               
 Shift <   w   x   c   v   b   n   ,   ;   :   !    Shift   
+------------------------------------------------------------------------+


 Here is the logical key numbers map:

+----------------------------------------------------------------+-------+
 41  2   3   4   5   6   7   8   9   10  11  12  13        
                                                           
+-------------------------------------------------------------------+----+
       16  17  18  19  20  21  22  23  24  25  26  27      
                                                           
+------+-------------------------------------------------------------+   
         30  31  32  33  34  35  36  37  38  39  40  43    
                                                           
+------+-------------------------------------------------------------+---
       86  44  45  46  47  48  49  50  51  52  53           
                                                            
+------+------------------------------------------------------+----------+
*/









//
// Checks if 'a' and 'b' are close on the same line on the current
// keyboard.
//
int areCloseOnKeyboard(unichar a,unichar b,struct keyboard* kbd) {
if (kbd==NULL) {
   // no keyboard means that this function is desactivated
   return 0;
}
if (a==NO_KEY_DEFINED || b==NO_KEY_DEFINED) {
   fprintf(stderr,"NO_KEY_DEFINED error in areCloseOnKeyboard\n");
   fatal_error(1);
}
unichar* tab[6];
unichar* key;
//----- upper line -----------
tab[0]=kbd->low0;
tab[1]=kbd->up0;
tab[2]=kbd->alt0;
for (int k=0;k<3;k++) {
   key=tab[k];
   // extrem left char
   if (key[0]==a) {
      return b==key[1];
   }
   // extrem right char
   if (key[PC_KEYBOARD_UPPER_LINE_SIZE-1]==a) {
      return b==key[PC_KEYBOARD_UPPER_LINE_SIZE-2];
   }
   // middle chars
   for (int i=1;i<PC_KEYBOARD_UPPER_LINE_SIZE-1;i++) {
      if (key[i]==a) {
         return b==key[i-1] || b==key[i+1];
      }
   }
}
//----- middle lines -----------
tab[0]=kbd->low1;
tab[1]=kbd->up1;
tab[2]=kbd->alt1;
tab[3]=kbd->low2;
tab[4]=kbd->up2;
tab[5]=kbd->alt2;
for (int k=0;k<6;k++) {
   key=tab[k];
   // extrem left char
   if (key[0]==a) {
      return b==key[1];
   }
   // extrem right char
   if (key[PC_KEYBOARD_LINE_SIZE-1]==a) {
      return b==key[PC_KEYBOARD_LINE_SIZE-2];
   }
   // middle chars
   for (int i=1;i<PC_KEYBOARD_LINE_SIZE-1;i++) {
      if (key[i]==a) {
         return b==key[i-1] || b==key[i+1];
      }
   }
}
//----- lower line -----------
tab[0]=kbd->low3;
tab[1]=kbd->up3;
tab[2]=kbd->alt3;
for (int k=0;k<3;k++) {
   key=tab[k];
   // extrem left char
   if (key[0]==a) {
      return b==key[1];
   }
   // extrem right char
   if (key[PC_KEYBOARD_LOWER_LINE_SIZE-1]==a) {
      return b==key[PC_KEYBOARD_LOWER_LINE_SIZE-2];
   }
   // middle chars
   for (int i=1;i<PC_KEYBOARD_LOWER_LINE_SIZE-1;i++) {
      if (key[i]==a) {
         return b==key[i-1] || b==key[i+1];
      }
   }
}
return 0;
}


//
// Checks if 'a' and 'b' are close on the same line on the current
// keyboard.
//
int areCloseOnKeyboard(unichar a,unichar b) {
return areCloseOnKeyboard(a,b,current_keyboard);
}


//-----------------------------------------


struct keyboard* new_keyboard() {
struct keyboard* k=(struct keyboard*)malloc(sizeof(struct keyboard));
if (k==NULL) {
   fprintf(stderr,"malloc error in new_keyboard\n");
}
return k;
}

void free_keyboard(struct keyboard* k) {
if (k==NULL) return;
free(k);
}




void init_keyboard(struct keyboard* keyboard) {
if (keyboard==NULL) {
   fprintf(stderr,"NULL error in init_keyboard\n");
   fatal_error(1);
}
strcpy(keyboard->name,"");
strcpy(keyboard->comment,"Empty keyboard");
for (int i=0;i<PC_KEYBOARD_UPPER_LINE_SIZE;i++) {
   keyboard->low0[i]=NO_KEY_DEFINED;
   keyboard->up0[i]=NO_KEY_DEFINED;
   keyboard->alt0[i]=NO_KEY_DEFINED;
}
for (int i=0;i<PC_KEYBOARD_LINE_SIZE;i++) {
   keyboard->low1[i]=NO_KEY_DEFINED;
   keyboard->up1[i]=NO_KEY_DEFINED;
   keyboard->alt1[i]=NO_KEY_DEFINED;
   keyboard->low2[i]=NO_KEY_DEFINED;
   keyboard->up2[i]=NO_KEY_DEFINED;
   keyboard->alt2[i]=NO_KEY_DEFINED;
}
for (int i=0;i<PC_KEYBOARD_LOWER_LINE_SIZE;i++) {
   keyboard->low3[i]=NO_KEY_DEFINED;
   keyboard->up3[i]=NO_KEY_DEFINED;
   keyboard->alt3[i]=NO_KEY_DEFINED;
}
}


void init_azerty_layout(struct keyboard* keyboard) {
init_keyboard(keyboard);
strcpy(keyboard->name,"azerty-layout");
strcpy(keyboard->comment,"Strict ANSI azerty layout");
u_strcpy_char_with_no_zero(keyboard->low1,"azertyuiop");
u_strcpy_char_with_no_zero(keyboard->low2,"qsdfghjklm");
u_strcpy_char_with_no_zero(&(keyboard->low3[1]),"wxcvbn");
u_strcpy_char_with_no_zero(keyboard->up1,"AZERTYUIOP");
u_strcpy_char_with_no_zero(keyboard->up2,"QSDFGHJKLM");
u_strcpy_char_with_no_zero(&(keyboard->up3[1]),"WXCVBN");
}


void init_qwerty_layout(struct keyboard* keyboard) {
init_keyboard(keyboard);
strcpy(keyboard->name,"qwerty-layout");
strcpy(keyboard->comment,"Strict ANSI qwerty layout");
u_strcpy_char_with_no_zero(keyboard->low1,"qwertyuiop");
u_strcpy_char_with_no_zero(keyboard->low2,"asdfghjkl");
u_strcpy_char_with_no_zero(&(keyboard->low3[1]),"zxcvbnm");
u_strcpy_char_with_no_zero(keyboard->up1,"QWERTYUIOP");
u_strcpy_char_with_no_zero(keyboard->up2,"ASDFGHJKL");
u_strcpy_char_with_no_zero(&(keyboard->up3[1]),"ZXCVBNM");
}


void init_qwertz_layout(struct keyboard* keyboard) {
init_keyboard(keyboard);
strcpy(keyboard->name,"qwertz-layout");
strcpy(keyboard->comment,"Strict ANSI qwertz layout");
u_strcpy_char_with_no_zero(keyboard->low1,"qwertzuiop");
u_strcpy_char_with_no_zero(keyboard->low2,"asdfghjkl");
u_strcpy_char_with_no_zero(&(keyboard->low3[1]),"yxcvbnm");
u_strcpy_char_with_no_zero(keyboard->up1,"QWERTZUIOP");
u_strcpy_char_with_no_zero(keyboard->up2,"ASDFGHJKL");
u_strcpy_char_with_no_zero(&(keyboard->up3[1]),"YXCVBNM");
}


void init_azerty(struct keyboard* keyboard) {
init_azerty_layout(keyboard);
strcpy(keyboard->name,"azerty");
strcpy(keyboard->comment,"Azerty layout extended with latin accented letters");
keyboard->low0[0]=U_SUPERSCRIPT_TWO;
keyboard->low0[1]='&';
keyboard->low0[2]=U_ECUTE;
keyboard->low0[3]='"';
keyboard->low0[4]='\'';
keyboard->low0[5]='(';
keyboard->low0[6]=U_SECTION_MARK;
keyboard->low0[7]=U_EGRAVE;
keyboard->low0[8]='!';
keyboard->low0[9]=U_CCEDILLA;
keyboard->low0[10]=U_AGRAVE;
keyboard->low0[11]=')';
keyboard->low0[12]='-';
//-----------------------------
u_strcpy_char_with_no_zero(&(keyboard->up0[1]),"1234567890");
keyboard->up0[11]=U_DEGREE;
keyboard->up0[12]='_';
//-----------------------------
keyboard->alt0[1]='|';
keyboard->alt0[2]='@';
keyboard->alt0[6]='^';
keyboard->alt0[9]='{';
keyboard->alt0[10]='}';
keyboard->alt0[12]='\\';
//------------------------
// we do not take into account dead keys, i.e. keys that
// must be combined with other keys to produce a symbol.
// Ex: the dead circumflex key must be combined with a letter to 
//     produce a letter, whereas the normal circumflex key prints
//     the '^' char.
//
keyboard->alt1[10]='[';
keyboard->low1[11]='$';
keyboard->up1[11]='*';
keyboard->alt1[11]=']';
//------------------------
keyboard->low2[10]=U_UGRAVE;
keyboard->up2[10]='%';
keyboard->alt2[10]=U_ACUTE;
keyboard->low2[11]=U_MU;
keyboard->up2[11]=U_STERLING;
//-------------------------
keyboard->low3[0]='<';
keyboard->up3[0]='>';
keyboard->alt3[0]='\\';
u_strcpy_char_with_no_zero(&(keyboard->low3[7]),",;:=");
u_strcpy_char_with_no_zero(&(keyboard->up3[7]),"?./+");
}



void init_fr_latin1(struct keyboard* keyboard) {
init_azerty_layout(keyboard);
strcpy(keyboard->name,"fr-latin1");
strcpy(keyboard->comment,"Keyboard used for French of France");
keyboard->low0[0]=U_SUPERSCRIPT_TWO;
keyboard->low0[1]='&';
keyboard->low0[2]=U_ECUTE;
keyboard->low0[3]='"';
keyboard->low0[4]='\'';
keyboard->low0[5]='(';
keyboard->low0[6]='-';
keyboard->low0[7]=U_EGRAVE;
keyboard->low0[8]='_';
keyboard->low0[9]=U_CCEDILLA;
keyboard->low0[10]=U_AGRAVE;
keyboard->low0[11]=')';
keyboard->low0[12]='=';
//-----------------------------
u_strcpy_char_with_no_zero(&(keyboard->up0[1]),"1234567890");
keyboard->up0[11]=U_DEGREE;
keyboard->up0[12]='+';
//-----------------------------
keyboard->alt0[2]=U_TILDE;
keyboard->alt0[3]=U_NUMBER_SIGN;
keyboard->alt0[4]='{';
keyboard->alt0[5]='[';
keyboard->alt0[6]='|';
keyboard->alt0[8]='\\';
keyboard->alt0[9]=U_CIRCUMFLEX;
keyboard->alt0[10]='@';
keyboard->alt0[11]=']';
keyboard->alt0[12]='}';
//------------------------
keyboard->alt1[2]=U_EURO;
keyboard->low1[11]='$';
keyboard->up1[11]=U_STERLING;
keyboard->alt1[11]=U_CURRENCY;
//------------------------
keyboard->low2[10]=U_UGRAVE;
keyboard->up2[10]='%';
keyboard->low2[11]='*';
keyboard->up2[11]=U_MU;
//-------------------------
keyboard->low3[0]='<';
keyboard->up3[0]='>';
u_strcpy_char_with_no_zero(&(keyboard->low3[7]),",;:!");
u_strcpy_char_with_no_zero(&(keyboard->up3[7]),"?./");
keyboard->up3[10]=U_SECTION_MARK;
}



void init_be_latin1(struct keyboard* keyboard) {
init_azerty_layout(keyboard);
strcpy(keyboard->name,"be-latin1");
strcpy(keyboard->comment,"Keyboard used for French of Belgium");
keyboard->low0[0]=U_SUPERSCRIPT_TWO;
keyboard->low0[1]='&';
keyboard->low0[2]=U_ECUTE;
keyboard->low0[3]='"';
keyboard->low0[4]='\'';
keyboard->low0[5]='(';
keyboard->low0[6]=U_SECTION_MARK;
keyboard->low0[7]=U_EGRAVE;
keyboard->low0[8]='!';
keyboard->low0[9]=U_CCEDILLA;
keyboard->low0[10]=U_AGRAVE;
keyboard->low0[11]=')';
keyboard->low0[12]='-';
//-----------------------------
keyboard->up0[0]=U_SUPERSCRIPT_THREE;
u_strcpy_char_with_no_zero(&(keyboard->up0[1]),"1234567890");
keyboard->up0[11]=U_DEGREE;
keyboard->up0[12]='_';
//-----------------------------
keyboard->alt0[1]='|';
keyboard->alt0[2]='@';
keyboard->alt0[3]='#';
keyboard->alt0[6]='^';
keyboard->alt0[9]='{';
keyboard->alt0[10]='}';
//------------------------
// we do not take into account dead keys, i.e. keys that
// must be combined with other keys to produce a symbol.
// Ex: the dead circumflex key must be combined with a letter to 
//     produce a letter, whereas the normal circumflex key prints
//     the '^' char.
//
keyboard->alt1[10]='[';
keyboard->low1[11]='$';
keyboard->up1[11]='*';
keyboard->alt1[11]=']';
//------------------------
keyboard->low2[10]=U_UGRAVE;
keyboard->up2[10]='%';
keyboard->low2[11]=U_MU;
keyboard->up2[11]=U_STERLING;
//-------------------------
keyboard->low3[0]='<';
keyboard->up3[0]='>';
keyboard->alt3[0]='\\';
u_strcpy_char_with_no_zero(&(keyboard->low3[7]),",;:=");
u_strcpy_char_with_no_zero(&(keyboard->up3[7]),"?./+");
}



void init_ansi_dvorak(struct keyboard* keyboard) {
init_keyboard(keyboard);
strcpy(keyboard->name,"ansi-dvorak");
strcpy(keyboard->comment,"Strict ANSI dvorak keyboard");
keyboard->low0[0]=U_GRAVE;
keyboard->up0[0]=U_TILDE;
u_strcpy_char_with_no_zero(&(keyboard->low0[1]),"1234567890[]");
u_strcpy_char_with_no_zero(&(keyboard->up0[1]), "!@#$%^&*(){}");
keyboard->alt0[2]='@';
keyboard->alt0[4]='$';
keyboard->alt0[7]='{';
keyboard->alt0[8]='[';
keyboard->alt0[9]=']';
keyboard->alt0[10]='}';
keyboard->alt0[12]=U_TILDE;
//----------------------------------
u_strcpy_char_with_no_zero(keyboard->low1,"',.pyfgcrl/=");
u_strcpy_char_with_no_zero(keyboard->up1, "\"<>PYFGCRL?+");
//----------------------------------
u_strcpy_char_with_no_zero(keyboard->low2,"aoeuidhtns-\\");
u_strcpy_char_with_no_zero(keyboard->up2, "AOEUIDHTNS_|");
keyboard->alt2[10]='\\';;
//----------------------------------
u_strcpy_char_with_no_zero(keyboard->low3,"<;qjkxbmwvz");
u_strcpy_char_with_no_zero(keyboard->up3, ">:QJKXBMWVZ");
keyboard->alt3[0]='|';;
}



void init_dvorak_l(struct keyboard* keyboard) {
init_keyboard(keyboard);
strcpy(keyboard->name,"dvorak-l");
strcpy(keyboard->comment,"ANSI left single-handed dvorak keyboard");
u_strcpy_char_with_no_zero(keyboard->low0,"`[]/pfmlj4321");
u_strcpy_char_with_no_zero(keyboard->up0, "~{}?PFMLJ$#@!");
//---------------------------------------
u_strcpy_char_with_no_zero(keyboard->low1,";qbyurso.65=");
u_strcpy_char_with_no_zero(keyboard->up1, ":QBYURSO>^%+");
//---------------------------------------
u_strcpy_char_with_no_zero(keyboard->low2,"<kcdtheaz87\\");
u_strcpy_char_with_no_zero(keyboard->up2, "_KCDTHEAZ*&|");
//---------------------------------------
u_strcpy_char_with_no_zero(&(keyboard->low3[1]),"'xgvwni,09");
u_strcpy_char_with_no_zero(&(keyboard->up3[1]),"\"XGVWNI<)(");
}


void init_dvorak_r(struct keyboard* keyboard) {
init_keyboard(keyboard);
strcpy(keyboard->name,"dvorak-r");
strcpy(keyboard->comment,"ANSI right single-handed dvorak keyboard");
u_strcpy_char_with_no_zero(keyboard->low0,"`1234jlmfp/[]");
u_strcpy_char_with_no_zero(keyboard->up0, "~!@#$JLMFP?{}");
//---------------------------------------
u_strcpy_char_with_no_zero(keyboard->low1,"56q.orsuyb;=");
u_strcpy_char_with_no_zero(keyboard->up1, "%^Q>ORSUYB:+");
//---------------------------------------
u_strcpy_char_with_no_zero(keyboard->low2,"78zaehtdck-\\");
u_strcpy_char_with_no_zero(keyboard->up2, "&*ZAEHTDCK_|");
//---------------------------------------
u_strcpy_char_with_no_zero(&(keyboard->low3[1]),"90x,inwvg'");
u_strcpy_char_with_no_zero(&(keyboard->up3[1]), "()X<INWVG\"");
}



void init_turkish_f(struct keyboard* keyboard) {
init_keyboard(keyboard);
strcpy(keyboard->name,"turkish-f");
strcpy(keyboard->comment,"Turkish F-keyboard");
u_strcpy_char_with_no_zero(keyboard->low0,"w1234567890xq");
u_strcpy_char_with_no_zero(keyboard->up0, "W!@\"'(/)_%:XQ");
u_strcpy_char_with_no_zero(&(keyboard->alt0[1]),"~&#$<>{}[]");
//-----------------------------
keyboard->low1[0]='f';
keyboard->low1[1]='g';
keyboard->low1[2]=360;
keyboard->low1[3]=375;
u_strcpy_char_with_no_zero(&(keyboard->low1[4]),"odrhp;^=");
keyboard->up1[0]='F';
keyboard->up1[1]='G';
keyboard->up1[2]=320;
u_strcpy_char_with_no_zero(&(keyboard->up1[3]),"IODRHP*`+");
keyboard->alt1[3]=356;
keyboard->alt1[4]=364;
//-----------------------------
u_strcpy_char_with_no_zero(keyboard->low2,"uiea");
keyboard->low2[4]=374;
u_strcpy_char_with_no_zero(&(keyboard->low2[5]),"tkmly");
keyboard->low2[10]=376;
keyboard->low2[11]='=';
keyboard->up2[0]='U';
keyboard->up2[1]=335;
u_strcpy_char_with_no_zero(&(keyboard->up2[2]),"EA");
keyboard->up2[4]=334;
u_strcpy_char_with_no_zero(&(keyboard->up2[5]),"TKMLY");
keyboard->up2[10]=336;
keyboard->up2[11]='+';
keyboard->alt2[0]=373;
keyboard->alt2[3]=342;
//-----------------------------
keyboard->low3[0]='\\';
keyboard->low3[1]='j';
keyboard->low3[2]=366;
keyboard->low3[3]='v';
keyboard->low3[4]='c';
keyboard->low3[5]=347;
u_strcpy_char_with_no_zero(&(keyboard->low3[6]),"zsb.,");
keyboard->up3[0]='|';
keyboard->up3[1]='J';
keyboard->up3[2]=326;
keyboard->up3[3]='V';
keyboard->up3[4]='C';
keyboard->up3[5]=307;
u_strcpy_char_with_no_zero(&(keyboard->up3[6]),"ZSB?-");
keyboard->alt3[10]='/';
}



void init_de(struct keyboard* keyboard) {
init_qwertz_layout(keyboard);
strcpy(keyboard->name,"de");
strcpy(keyboard->comment,"German keyboard with no non-latin letters");
keyboard->alt1[0]='@';
keyboard->alt3[7]=U_MU;
keyboard->alt1[2]=U_CURRENCY;
keyboard->alt3[3]=U_CENT;
u_strcpy_char_with_no_zero(keyboard->low0,"^1234567890\\'");
u_strcpy_char_with_no_zero(keyboard->up0, "~!\"#$%&/()=?`");
keyboard->alt0[2]=U_SUPERSCRIPT_TWO;
keyboard->alt0[3]=U_SUPERSCRIPT_THREE;
keyboard->alt0[8]='{';
keyboard->alt0[9]='[';
keyboard->alt0[10]=']';
keyboard->alt0[11]='}';
//
keyboard->low1[10]='@';
keyboard->low1[11]='+';
keyboard->up1[10]='\\';
keyboard->up1[11]='*';
keyboard->alt1[11]='~';
//
keyboard->low2[9]='[';
keyboard->up2[9]='{';
keyboard->low2[10]=']';
keyboard->up2[10]='}';
keyboard->low2[11]='#';
keyboard->up2[11]='\'';
//
keyboard->low3[0]='<';
keyboard->up3[0]='>';
keyboard->alt3[0]='|';
u_strcpy_char_with_no_zero(&(keyboard->low3[8]),",.-");
u_strcpy_char_with_no_zero(&(keyboard->up3[8]), "';:");
}



void init_de_latin1(struct keyboard* keyboard) {
init_qwertz_layout(keyboard);
strcpy(keyboard->name,"de-latin1");
strcpy(keyboard->comment,"German keyboard");
keyboard->alt1[0]='@';
keyboard->alt3[7]=U_MU;
keyboard->alt1[2]=U_CURRENCY;
keyboard->alt3[3]=U_CENT;
u_strcpy_char_with_no_zero(&(keyboard->low0[1]),"1234567890");
keyboard->low0[11]=U_SSHARP;
u_strcpy_char_with_no_zero(keyboard->up0,"!\"");
keyboard->up0[3]=U_SECTION_MARK;
u_strcpy_char_with_no_zero(&(keyboard->up0[4]),"$%&/()=?");
keyboard->alt0[2]=U_SUPERSCRIPT_TWO;
keyboard->alt0[3]=U_SUPERSCRIPT_THREE;
keyboard->alt0[7]='{';
keyboard->alt0[8]='[';
keyboard->alt0[9]=']';
keyboard->alt0[10]='}';
keyboard->alt0[11]='\\';
//
keyboard->low1[10]=U_UDIAERESIS;
keyboard->up1[10]=U_UDIAERESIS_CAP;
keyboard->low1[11]='+';
keyboard->up1[11]='*';
//
keyboard->low2[9]=U_ODIAERESIS;
keyboard->up2[9]=U_ODIAERESIS_CAP;
keyboard->low2[10]=U_ADIAERESIS;
keyboard->up2[10]=U_ADIAERESIS_CAP;
keyboard->low2[11]='#';
keyboard->up2[11]='\'';
//
keyboard->low3[0]='<';
keyboard->up3[0]='>';
keyboard->alt3[0]='|';
keyboard->low3[8]=',';
keyboard->low3[9]='.';
keyboard->low3[10]='-';
keyboard->up3[8]=';';
keyboard->up3[9]=':';
keyboard->up3[10]='_';
}





int init_keyboards() {
init_fr_latin1(&(keyboards[0]));
init_be_latin1(&(keyboards[1]));
init_ansi_dvorak(&(keyboards[2]));
init_dvorak_l(&(keyboards[3]));
init_dvorak_r(&(keyboards[4]));
init_turkish_f(&(keyboards[5]));
init_de_latin1(&(keyboards[6]));
return 0;
}


// this line is used to force the initialization of the keyboards
int phantom_var=init_keyboards();



struct keyboard* get_keyboard(char* name) {
for (int i=0;i<N_KEYBOARDS;i++) {
   if (!strcmp(name,keyboards[i].name)) {
      return &(keyboards[i]);
   }
}
return NULL;
}




void display_all_keyboards(FILE* out) {
for (int i=0;i<N_KEYBOARDS;i++) {
   display_keyboard(&(keyboards[i]),out);
   u_fprints_char("\n\n--------------------------------------------------------------\n\n",out);
}
}




//
// Displays the letter keys of a given keyboard
//
void display_keyboard(struct keyboard* keyboard,FILE* out) {
if (keyboard==NULL) {
   fprintf(stderr,"NULL keyboard error in display_keyboard\n");
   fatal_error(1);
}
u_fprints_char("      Keyboard name=",out);
u_fprints_char(keyboard->name,out);
u_fprints_char("\n\n      ",out);
u_fprints_char(keyboard->comment,out);
u_fprints_char("\n\n",out);
//----- first keyboard line -------
u_fputc(0x250C,out);
for (int i=0;i<PC_KEYBOARD_UPPER_LINE_SIZE;i++) {
   u_fputc(0x2500,out);
   u_fputc(0x2500,out);
   u_fputc(0x2500,out);
   u_fputc(0x2500,out);
   u_fputc(0x252C,out);
}
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2510,out);
u_fputc('\n',out);
//
u_fputc(0x2502,out);
for (int i=0;i<PC_KEYBOARD_UPPER_LINE_SIZE;i++) {
   u_fputc(' ',out);
   u_fputc((keyboard->up0[i]!=NO_KEY_DEFINED)?keyboard->up0[i]:' ',out);
   u_fputc(' ',out);
   u_fputc(' ',out);
   u_fputc(0x2502,out);
}
u_fprints_char(" <",out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fprints_char("  ",out);
u_fputc(0x2502,out);
u_fputc('\n',out);
//
u_fputc(0x2502,out);
for (int i=0;i<PC_KEYBOARD_UPPER_LINE_SIZE;i++) {
   u_fputc(' ',out);
   u_fputc((keyboard->low0[i]!=NO_KEY_DEFINED)?keyboard->low0[i]:' ',out);
   u_fputc(' ',out);
   u_fputc((keyboard->alt0[i]!=NO_KEY_DEFINED)?keyboard->alt0[i]:' ',out);
   u_fputc(0x2502,out);
}
u_fprints_char("       ",out);
u_fputc(0x2502,out);
u_fputc('\n',out);
//-------- second keyboard line -----------
u_fputc(0x251C,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2534,out);
for (int i=0;i<PC_KEYBOARD_LINE_SIZE;i++) {
   u_fputc(0x2500,out);
   u_fputc(0x252C,out);
   u_fputc(0x2500,out);
   u_fputc(0x2500,out);
   u_fputc(0x2534,out);
}
u_fputc(0x2500,out);
u_fputc(0x252C,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2524,out);
u_fputc('\n',out);
//
u_fputc(0x2502,out);
u_fprints_char(" |<-  ",out);
u_fputc(0x2502,out);
for (int i=0;i<PC_KEYBOARD_LINE_SIZE;i++) {
   u_fputc(' ',out);
   u_fputc((keyboard->up1[i]!=NO_KEY_DEFINED)?keyboard->up1[i]:' ',out);
   u_fputc(' ',out);
   u_fputc(' ',out);
   u_fputc(0x2502,out);
}
u_fprints_char("   ",out);
u_fputc(0x2502,out);
u_fprints_char(" ",out);
u_fputc(0x2502,out);
u_fputc('\n',out);
//
u_fputc(0x2502,out);
u_fprints_char(" ->|  ",out);
u_fputc(0x2502,out);
for (int i=0;i<PC_KEYBOARD_LINE_SIZE;i++) {
   u_fputc(' ',out);
   u_fputc((keyboard->low1[i]!=NO_KEY_DEFINED)?keyboard->low1[i]:' ',out);
   u_fputc(' ',out);
   u_fputc((keyboard->alt1[i]!=NO_KEY_DEFINED)?keyboard->alt1[i]:' ',out);
   u_fputc(0x2502,out);
}
u_fprints_char(" <",out);
u_fputc(0x2500,out);
u_fputc(0x2518,out);
u_fputc(' ',out);
u_fputc(0x2502,out);
u_fputc('\n',out);
//-------- third keyboard line --------
u_fputc(0x251C,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2534,out);
for (int i=0;i<PC_KEYBOARD_LINE_SIZE;i++) {
   u_fputc(0x2500,out);
   u_fputc(0x252C,out);
   u_fputc(0x2500,out);
   u_fputc(0x2500,out);
   u_fputc(0x2534,out);
}
u_fputc(0x2500,out);
u_fputc(0x2510,out);
u_fputc(' ',out);
u_fputc(' ',out);
u_fputc(' ',out);
u_fputc(0x2502,out);
u_fputc('\n',out);
//
u_fputc(0x2502,out);
u_fprints_char(" Caps   ",out);
u_fputc(0x2502,out);
for (int i=0;i<PC_KEYBOARD_LINE_SIZE;i++) {
   u_fputc(' ',out);
   u_fputc((keyboard->up2[i]!=NO_KEY_DEFINED)?keyboard->up2[i]:' ',out);
   u_fputc(' ',out);
   u_fputc(' ',out);
   u_fputc(0x2502,out);
}
u_fprints_char("   ",out);
u_fputc(0x2502,out);
u_fputc('\n',out);
//
u_fputc(0x2502,out);
u_fprints_char(" Lock   ",out);
u_fputc(0x2502,out);
for (int i=0;i<PC_KEYBOARD_LINE_SIZE;i++) {
   u_fputc(' ',out);
   u_fputc((keyboard->low2[i]!=NO_KEY_DEFINED)?keyboard->low2[i]:' ',out);
   u_fputc(' ',out);
   u_fputc((keyboard->alt2[i]!=NO_KEY_DEFINED)?keyboard->alt2[i]:' ',out);
   u_fputc(0x2502,out);
}
u_fprints_char("   ",out);
u_fputc(0x2502,out);
u_fputc('\n',out);
//-------- fourth keyboard line ----------
u_fputc(0x251C,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x252C,out);
u_fputc(0x2500,out);
u_fputc(0x2534,out);
for (int i=0;i<PC_KEYBOARD_LOWER_LINE_SIZE;i++) {
   u_fputc(0x2500,out);
   u_fputc(0x2500,out);
   u_fputc(0x252C,out);
   u_fputc(0x2500,out);
   u_fputc(0x2534,out);
}
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2534,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2524,out);
u_fputc('\n',out);
//
u_fputc(0x2502,out);
u_fprints_char("      ",out);
u_fputc(0x2502,out);
for (int i=0;i<PC_KEYBOARD_LOWER_LINE_SIZE;i++) {
   u_fputc(' ',out);
   u_fputc((keyboard->up3[i]!=NO_KEY_DEFINED)?keyboard->up3[i]:' ',out);
   u_fputc(' ',out);
   u_fputc(' ',out);
   u_fputc(0x2502,out);
}
u_fprints_char("          ",out);
u_fputc(0x2502,out);
u_fputc('\n',out);
//
u_fputc(0x2502,out);
u_fprints_char(" Shift",out);
u_fputc(0x2502,out);
for (int i=0;i<PC_KEYBOARD_LOWER_LINE_SIZE;i++) {
   u_fputc(' ',out);
   u_fputc((keyboard->low3[i]!=NO_KEY_DEFINED)?keyboard->low3[i]:' ',out);
   u_fputc(' ',out);
   u_fputc((keyboard->alt3[i]!=NO_KEY_DEFINED)?keyboard->alt3[i]:' ',out);
   u_fputc(0x2502,out);
}
u_fprints_char("  Shift   ",out);
u_fputc(0x2502,out);
u_fputc('\n',out);
//
u_fputc(0x2514,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2534,out);
for (int i=0;i<PC_KEYBOARD_LOWER_LINE_SIZE-1;i++) {
   u_fputc(0x2500,out);
   u_fputc(0x2500,out);
   u_fputc(0x2500,out);
   u_fputc(0x2500,out);
   u_fputc(0x2534,out);
}
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2534,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2500,out);
u_fputc(0x2518,out);
u_fputc('\n',out);
}
